"""
.. module:: session
   :synopsis: Main module for oct2py package.
              Contains the Octave session object Oct2Py

.. moduleauthor:: Steven Silvester <steven.silvester@ieee.org>

"""
from __future__ import print_function
import os
import re
import atexit
import doctest
import subprocess
import sys
import threading
import time


pexpect = None
if not os.name == 'nt':
    # needed for testing support
    if not hasattr(sys.stdout, 'buffer'):  # pragma: no cover
        class Dummy(object):

            def write(self):
                pass
        try:
            sys.stdout.buffer = Dummy()
        except AttributeError:
            pass
    try:
        import pexpect
    except ImportError:
        pass

from .matwrite import MatWrite
from .matread import MatRead
from .utils import get_nout, Oct2PyError, get_log
from .compat import unicode, PY2, queue


class Oct2Py(object):

    """Manages an Octave session.

    Uses MAT files to pass data between Octave and Numpy.
    The function must either exist as an m-file in this directory or
    on Octave's path.
    The first command will take about 0.5s for Octave to load up.
    The subsequent commands will be much faster.

    You may provide a logger object for logging events, or the oct2py.get_log()
    default will be used.  Events will be logged as debug unless verbose is set
    when calling a command, then they will be logged as info.

    """

    def __init__(self, logger=None, timeout=-1):
        """Start Octave and create our MAT helpers
        """
        self.timeout = timeout
        if not logger is None:
            self.logger = logger
        else:
            self.logger = get_log()
        self.restart()

    def __enter__(self):
        """Return octave object, restart session if necessary"""
        if not self._session:
            self.restart()
        return self

    def __exit__(self, type, value, traceback):
        """Close session"""
        self.close()

    def close(self):
        """Closes this octave session and removes temp files
        """
        if self._session:
            self._session.close()
        self._session = None
        self._writer.remove_file()
        self._reader.remove_file()

    def run(self, script, **kwargs):
        """
        Run artibrary Octave code.

        Parameters
        -----------
        script : str
            Command script to send to Octave for execution.
        verbose : bool, optional
            Log Octave output at info level.

        Returns
        -------
        out : str
            Octave printed output.

        Raises
        ------
        Oct2PyError
            If the script cannot be run by Octave.

        Examples
        --------
        >>> from oct2py import octave
        >>> out = octave.run('y=ones(3,3)')
        >>> print(out)
        y =
        <BLANKLINE>
                1        1        1
                1        1        1
                1        1        1
        <BLANKLINE>
        >>> octave.run('x = mean([[1, 2], [3, 4]])')
        u'x =  2.5000'

        """
        # don't return a value from a script
        kwargs['nout'] = 0
        return self.call(script, **kwargs)

    def call(self, func, *inputs, **kwargs):
        """
        Call an Octave function with optional arguments.

        Parameters
        ----------
        func : str
            Function name to call.
        inputs : array_like
            Variables to pass to the function.
        nout : int, optional
            Number of output arguments.
            This is set automatically based on the number of
            return values requested (see example below).
            You can override this behavior by passing a
            different value.
        verbose : bool, optional
             Log Octave output at info level.

        Returns
        -------
        out : str or tuple
            If nout > 0, returns the values from Octave as a tuple.
            Otherwise, returns the output displayed by Octave.

        Raises
        ------
        Oct2PyError
            If the call is unsucessful.

        Examples
        --------
        >>> from oct2py import octave
        >>> b = octave.call('ones', 1, 2)
        >>> print(b)
        [[ 1.  1.]]
        >>> x, y = 1, 2
        >>> a = octave.call('zeros', x, y)
        >>> a
        array([[ 0.,  0.]])
        >>> U, S, V = octave.call('svd', [[1, 2], [1, 3]])
        >>> print((U, S, V))
        (array([[-0.57604844, -0.81741556],
               [-0.81741556,  0.57604844]]), array([[ 3.86432845,  0.        ],
               [ 0.        ,  0.25877718]]), array([[-0.36059668, -0.93272184],
               [-0.93272184,  0.36059668]]))

        """
        if self._first_run:
            self._first_run = False
            self._setup_session()

        verbose = kwargs.get('verbose', False)
        nout = kwargs.get('nout', get_nout())
        timeout = kwargs.get('timeout', self.timeout)

        # handle references to script names - and paths to them
        if func.endswith('.m'):
            if os.path.dirname(func):
                self.addpath(os.path.dirname(func))
                func = os.path.basename(func)
            func = func[:-2]

        # these three lines will form the commands sent to Octave
        # load("-v6", "infile", "invar1", ...)
        # [a, b, c] = foo(A, B, C)
        # save("-v6", "outfile", "outvar1", ...)
        load_line = call_line = save_line = ''

        if nout:
            # create a dummy list of var names ("a", "b", "c", ...)
            # use ascii char codes so we can increment
            argout_list, save_line = self._reader.setup(nout)
            call_line = '[{0}] = '.format(', '.join(argout_list))
        if inputs:
            argin_list, load_line = self._writer.create_file(inputs)
            call_line += '{0}({1})'.format(func, ', '.join(argin_list))
        elif nout:
            # call foo() - no arguments
            call_line += '{0}()'.format(func)
        else:
            # run foo
            call_line += '{0}'.format(func)

        pre_call = '\nglobal __oct2py_figures = [];\n'
        post_call = ''

        if 'command' in kwargs and not '__ipy_figures' in func:
            if not call_line.endswith(')') and nout:
                call_line += '();\n'
            else:
                call_line += ';\n'
            post_call += """
            # Save output of the last execution
                if exist("ans") == 1
                  _ = ans;
                else
                  _ = "__no_answer";
                end
            """

        # do not interfere with octavemagic logic
        if not "DefaultFigureCreateFcn" in call_line:
            post_call += """
            for f = __oct2py_figures
                try
                   refresh(f);
                end
            end"""

        # create the command and execute in octave
        cmd = [load_line, pre_call, call_line, post_call, save_line]
        resp = self._eval(cmd, verbose=verbose, timeout=timeout)

        if nout:
            return self._reader.extract_file(argout_list)
        elif 'command' in kwargs:
            try:
                ans = self.get('_')
            except (KeyError, Oct2PyError):
                return
            # Unfortunately, Octave doesn't have a "None" object,
            # so we can't return any NaN outputs
            if isinstance(ans, (str, unicode)) and ans == "__no_answer":
                ans = None
            return ans
        else:
            return resp

    def put(self, names, var, verbose=False, timeout=-1):
        """
        Put a variable into the Octave session.

        Parameters
        ----------
        names : str or list
            Name of the variable(s).
        var : object or list
            The value(s) to pass.
        timeout : float
            Time to wait for response from Octave (per character).

        Examples
        --------
        >>> from oct2py import octave
        >>> y = [1, 2]
        >>> octave.put('y', y)
        >>> octave.get('y')
        array([[1, 2]])
        >>> octave.put(['x', 'y'], ['spam', [1, 2, 3, 4]])
        >>> octave.get(['x', 'y'])
        (u'spam', array([[1, 2, 3, 4]]))

        """
        if isinstance(names, str):
            var = [var]
            names = [names]
        for name in names:
            if name.startswith('_'):
                raise Oct2PyError('Invalid name {0}'.format(name))
        _, load_line = self._writer.create_file(var, names)
        self._eval(load_line, verbose=verbose, timeout=timeout)

    def get(self, var, verbose=False, timeout=-1):
        """
        Retrieve a value from the Octave session.

        Parameters
        ----------
        var : str
            Name of the variable to retrieve.
        timeout : float
            Time to wait for response from Octave (per character).

        Returns
        -------
        out : object
            Object returned by Octave.

        Raises:
          Oct2PyError
            If the variable does not exist in the Octave session.

        Examples:
          >>> from oct2py import octave
          >>> y = [1, 2]
          >>> octave.put('y', y)
          >>> octave.get('y')
          array([[1, 2]])
          >>> octave.put(['x', 'y'], ['spam', [1, 2, 3, 4]])
          >>> octave.get(['x', 'y'])
          (u'spam', array([[1, 2, 3, 4]]))

        """
        if isinstance(var, str):
            var = [var]
        # make sure the variable(s) exist
        for variable in var:
            if self._eval("exist {0}".format(variable),
                          verbose=False) == 'ans = 0' and not variable == '_':
                raise Oct2PyError('{0} does not exist'.format(variable))
        argout_list, save_line = self._reader.setup(len(var), var)
        self._eval(save_line, verbose=verbose, timeout=timeout)
        return self._reader.extract_file(argout_list)

    def lookfor(self, string, verbose=False, timeout=-1):
        """
        Call the Octave "lookfor" command.

        Uses with the "-all" switch to search within help strings.

        Parameters
        ----------
        string : str
            Search string for the lookfor command.
        verbose : bool, optional
             Log Octave output at info level.
        timeout : float
            Time to wait for response from Octave (per character).

        Returns
        -------
        out : str
            Output from the Octave lookfor command.

        """
        return self.run('lookfor -all {0}'.format(string), verbose=verbose,
                        timeout=timeout)

    def _eval(self, cmds, verbose=True, log=True, timeout=-1):
        """
        Perform raw Octave command.

        This is a low-level command, and should not technically be used
        directly.  The API could change. You have been warned.

        Parameters
        ----------
        cmds : str or list
            Commands(s) to pass directly to Octave.
        verbose : bool, optional
             Log Octave output at info level.
        timeout : float
            Time to wait for response from Octave (per character).

        Returns
        -------
        out : str
            Results printed by Octave.

        Raises
        ------
        Oct2PyError
            If the command(s) fail.

        """
        if not self._session:
            raise Oct2PyError('No Octave Session')
        if isinstance(cmds, str):
            cmds = [cmds]
        if verbose and log:
            [self.logger.info(line) for line in cmds]
        elif log:
            [self.logger.debug(line) for line in cmds]
        if timeout == -1:
            timeout = self.timeout
        return self._session.evaluate(cmds, verbose, log, self.logger,
                                      timeout=timeout)

    def _make_octave_command(self, name, doc=None):
        """Create a wrapper to an Octave procedure or object

        Adapted from the mlabwrap project

        """
        def octave_command(*args, **kwargs):
            """ Octave command """
            kwargs['nout'] = get_nout()
            kwargs['verbose'] = kwargs.get('verbose', False)
            if not 'Built-in Function' in doc:
                self._eval('clear {}'.format(name), log=False, verbose=False)
            kwargs['command'] = True
            return self.call(name, *args, **kwargs)
        # convert to ascii for pydoc
        try:
            doc = doc.encode('ascii', 'replace').decode('ascii')
        except UnicodeDecodeError:
            pass
        octave_command.__doc__ = "\n" + doc
        octave_command.__name__ = name
        return octave_command

    def _get_doc(self, name):
        """
        Get the documentation of an Octave procedure or object.

        Parameters
        ----------
        name : str
            Function name to search for.

        Returns
        -------
        out : str
          Documentation string.

        Raises
        ------
        Oct2PyError
           If the procedure or object does not exist.

        """
        if name == 'keyboard':
            return 'Built-in Function: keyboard ()'
        exist = self._eval('exist {0}'.format(name), log=False, verbose=False)
        if exist.strip() == 'ans = 0':
            msg = 'Name: "%s" does not exist on the Octave session path'
            raise Oct2PyError(msg % name)
        doc = 'No documentation for %s' % name
        try:
            doc = self._eval('help {0}'.format(name), log=False, verbose=False)
        except Oct2PyError as e:
            if 'syntax error' in str(e):
                raise(e)
            try:
                doc = self._eval('type {0}'.format(name), log=False,
                                 verbose=False)
                doc = '\n'.join(doc.splitlines()[:3])
            except Oct2PyError as e:
                pass
        return doc

    def __getattr__(self, attr):
        """Automatically creates a wapper to an Octave function or object.

        Adapted from the mlabwrap project.

        """
        # needed for help(Oct2Py())
        if attr == '__name__':
            return super(Oct2Py, self).__getattr__(attr)
        elif attr == '__file__':
            return __file__
        # close_ -> close
        if attr[-1] == "_":
            name = attr[:-1]
        else:
            name = attr
        doc = self._get_doc(name)
        octave_command = self._make_octave_command(name, doc)
        #!!! attr, *not* name, because we might have python keyword name!
        setattr(self, attr, octave_command)
        return octave_command

    def _setup_session(self):
        try:
            self._eval("graphics_toolkit('gnuplot')", verbose=False)
        except Oct2PyError:  # pragma: no cover
            pass
        # set up the plot renderer
        self.run("""
            global __oct2py_figures = [];
            page_screen_output(0);

            function fig_create(src, event)
              global __oct2py_figures;
              __oct2py_figures(size(__oct2py_figures) + 1) = src;
            end

            set(0, 'DefaultFigureCreateFcn', @fig_create);
        """)

    def restart(self):
        """Restart an Octave session in a clean state
        """
        self._session = _Session()
        self._first_run = True
        self._reader = MatRead()
        self._writer = MatWrite()


class _Reader(object):

    """Read characters from an Octave session in a thread.
    """

    def __init__(self, proc, queue):
        self.proc = proc
        self.queue = queue
        self.thread = threading.Thread(target=self.read_incoming)
        self.thread.setDaemon(True)
        self.thread.start()

    def read_incoming(self):
        while 1:
            char = self.proc.stdout.read(1)
            try:
                self.queue.put(char)
            except:
                return


class _Session(object):

    """Low-level session Octave session interaction.
    """

    def __init__(self):
        self.timeout = int(1e6)
        self.use_pexpect = not pexpect is None
        self.read_queue = queue.Queue()
        self.interaction_file = '__oct2py_interact_%s' % id(self)
        self.proc = self.start()
        self.stdout = sys.stdout
        self.set_timeout()
        atexit.register(self.close)

    def start(self):
        """
        Start an octave session in a subprocess.

        Returns
        =======
        out : fid
            File descriptor for the Octave subprocess

        Raises
        ======
        Oct2PyError
            If the session is not opened sucessfully.

        Notes
        =====
        Options sent to Octave: -q is quiet startup, --braindead is
        Matlab compatibilty mode.

        """
        if self.use_pexpect:
            try:
                return pexpect.spawn('octave', ['-q', '--braindead'],
                                     timeout=60.)
            except Exception:
                return self.start_subprocess()
        else:
            return self.start_subprocess()

    def start_subprocess(self):
        """Start octave using a subprocess (no tty support)"""
        self.use_pexpect = False
        errmsg = ('\n\nPlease install GNU Octave and put it in your path\n')
        ON_POSIX = 'posix' in sys.builtin_module_names
        kwargs = dict(stderr=subprocess.STDOUT, stdin=subprocess.PIPE,
                      stdout=subprocess.PIPE, close_fds=ON_POSIX,
                      bufsize=0)
        if os.name == 'nt':
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            kwargs['startupinfo'] = startupinfo
        try:
            proc = subprocess.Popen(['octave', '-q', '--braindead'], **kwargs)
        except OSError:  # pragma: no cover
            raise Oct2PyError(errmsg)
        else:
            self.reader = _Reader(proc, self.read_queue)
            return proc

    def set_timeout(self, timeout=-1):
        if timeout == -1:
            timeout = int(1e6)
        if self.use_pexpect:
            self.proc.timeout = timeout
        else:
            self.timeout = timeout

    def evaluate(self, cmds, verbose=True, log=True, logger=None, timeout=-1):
        """Perform the low-level interaction with an Octave Session
        """
        if not timeout == -1:
            self.set_timeout(timeout)
        if not self.proc:
            raise Oct2PyError('Session Closed, try a restart()')
        # use ascii code 21 to signal an error and 3
        # to signal end of text
        lines = ['try', 'disp(char(3))', '\n'.join(cmds), 'disp(char(3))',
                 'catch', 'disp(lasterr())', 'disp(char(21))',
                 'end', '']
        if len(cmds) == 5:
            main_line = cmds[2].strip()
        else:
            main_line = '\n'.join(cmds)
        output = '\n'.join(lines)
        self.write(output)
        resp = self.expect(['\x03', 'syntax error'])
        if resp.endswith('syntax error'):
            if self.use_pexpect:
                resp += self.expect('\^')
                self.expect('\^')
                self.expect('\^')
                self.expect('\^')
            else:
                resp += self.expect('^')
            resp += self.expect('\n')
            self.handle_syntax_error(resp, main_line)
            return
        if not self.use_pexpect:
            try:
                self.proc.stdin.flush()
            except OSError:  # pragma: no cover
                pass
        resp = []
        syntax_error = False
        while 1:
            line = self.expect(['\n', '\A[\w]+>>? '])
            if line.rstrip().endswith('\x03'):
                break
            elif line.rstrip().endswith('\x15'):
                msg = ('Oct2Py tried to run:\n"""\n{0}\n"""\n'
                       'Octave returned:\n{1}'
                       .format(main_line, '\n'.join(resp)))
                raise Oct2PyError(msg)
            elif not line.endswith('\n') and not syntax_error:
                self.interact(line)
                self.write('clear _\n')
                resp = resp[:-4]
                self.expect('\x03')
                continue
            line = line.rstrip()
            if "syntax error" in line:
                syntax_error = True
            elif syntax_error and "^" in line:
                resp.append(line)
                self.handle_syntax_error(''.join(resp), main_line)
                return
            if verbose and logger:
                logger.info(line)
            elif log and logger:
                logger.debug(line)
            if resp or line:
                resp.append(line)
        return '\n'.join(resp)

    def handle_syntax_error(self, resp, main_line):
        """Handle an Octave syntax error"""
        errline = '\n'.join(resp.splitlines()[-2:])
        msg = ('Oct2Py tried to run:\n"""\n%s\n"""\n'
               'Octave returned Syntax Error:\n""""\n%s\n"""' % (main_line,
                                                      errline))
        msg += '\nIf using an m-file script, make sure it runs in Octave'
        if not self.use_pexpect:
            msg += '\nSession Closed by Octave'
            self.close()
            raise Oct2PyError(msg)
        else:
            raise Oct2PyError(msg)

    def expect(self, strings):
        """Look for a string or strings in the incoming data"""
        if not isinstance(strings, list):
            strings = [strings]
        line = ''
        if self.use_pexpect:
            try:
                self.proc.expect(strings)
            except pexpect.TIMEOUT:
                self.close()
                raise Oct2PyError('Session Timed Out, closing')
            line = self.proc.before + self.proc.after
            try:
                return line.decode('utf-8', 'replace')
            except:
                return line
        else:
            line = ''
            while 1:
                line += self.read()
                for string in strings:
                    if re.search(string, line):
                        return line

    def find_prompt(self, prompt='debug> ', disp=True):
        """Look for the prompt in the Octave output, print chars if disp"""
        output = ''
        while 1:
            output += self.read()
            if output.endswith(prompt):
                if disp:
                    self.stdout.write(output)
                return

    def read(self, n=1):
        """Read characters from the process with utf-8 encoding"""
        if self.use_pexpect:
            chars = self.proc.read(n)
            try:
                return chars.decode('utf-8', 'replace')
            except:
                return chars
        else:
            t0 = time.time()
            chars = []
            while 1:
                try:
                    chars.append(self.read_queue.get_nowait())
                except queue.Empty:
                    pass
                time.sleep(1e-6)
                if len(chars) == n:
                    chars = b''.join(chars)
                    return chars.decode('utf-8', 'replace')
                if (time.time() - t0) > self.timeout:
                    self.close()
                    raise Oct2PyError('Session Timed Out, closing')

    def write(self, message):
        """Write a message to the process using utf-8 encoding"""
        if self.use_pexpect:
            self.proc.write(message.encode('utf-8'))
        else:
            self.proc.stdin.write(message.encode('utf-8'))

    def interact(self, prompt='debug> '):
        """Manage an Octave Debug Prompt interaction"""
        msg = 'Entering Octave Debug Prompt...\n%s' % prompt
        self.stdout.write(msg)
        while 1:
            inp_func = input if not PY2 else raw_input
            try:
                inp = inp_func() + '\n'
            except EOFError:
                return
            if inp in ['exit\n', 'quit\n', 'dbcont\n', 'dbquit\n']:
                inp = 'return\n'
            self.write('disp(char(3));' + inp)
            if inp == 'return\n':
                return
            self.expect('\x03\r?\n')
            self.find_prompt(prompt)

    def close(self):
        """Cleanly close an Octave session
        """
        try:
            self.write('exit\n')
        except (IOError, AttributeError):
            pass
        try:
            self.proc.terminate()
        except (OSError, AttributeError):  # pragma: no cover
            pass
        self.proc = None


def _test():  # pragma: no cover
    """Run the doctests for this module.
    """
    print('Starting doctest')
    doctest.testmod()
    print('Completed doctest')


if __name__ == "__main__":  # pragma: no cover
    _test()



"""
oct2py_test - Test value passing between python and Octave.

Known limitations
-----------------
* The following Numpy array types cannot be sent directly via a MAT file.  The
float16/96/128 and complex192/256 can be recast as float64 and complex128.
   ** float16('e')
   ** float96('g')
   ** float128
   ** complex192('G')
   ** complex256
   ** read-write buffer('V')
"""
from __future__ import absolute_import, print_function
import logging
import os
import pickle
import sys

import numpy as np
import numpy.testing as test


import oct2py
from oct2py import Oct2Py, Oct2PyError
from oct2py.utils import Struct
from oct2py.compat import unicode, long, StringIO


octave = Oct2Py()
octave.addpath(os.path.dirname(__file__))
DATA = octave.test_datatypes()


TYPE_CONVERSIONS = [(int, 'int32', np.int32),
                (long, 'int64', np.int64),
                (float, 'double', np.float64),
                (complex, 'double', np.complex128),
                (str, 'char', unicode),
                (unicode, 'cell', unicode),
                (bool, 'int8', np.int8),
                (None, 'double', np.float64),
                (dict, 'struct', Struct),
                (np.int8, 'int8', np.int8),
                (np.int16, 'int16', np.int16),
                (np.int32, 'int32', np.int32),
                (np.int64, 'int64', np.int64),
                (np.uint8, 'uint8', np.uint8),
                (np.uint16, 'uint16', np.uint16),
                (np.uint32, 'uint32', np.uint32),
                (np.uint64, 'uint64', np.uint64),
                #(np.float16, 'double', np.float64),
                (np.float32, 'double', np.float64),
                (np.float64, 'double', np.float64),
                (np.str, 'char', np.unicode),
                (np.double, 'double', np.float64),
                (np.complex64, 'double', np.complex128),
                (np.complex128, 'double', np.complex128), ]


def test_python_conversions():
    """Test roundtrip python type conversions
    """
    oc = Oct2Py()
    oc.addpath(os.path.dirname(__file__))
    for out_type, oct_type, in_type in TYPE_CONVERSIONS:
        if out_type == dict:
            outgoing = dict(x=1)
        elif out_type == None:
            outgoing = None
        else:
            outgoing = out_type(1)
        incoming, octave_type = oc.roundtrip(outgoing)
        if octave_type == 'int32' and oct_type == 'int64':
            pass
        elif octave_type == 'char' and oct_type == 'cell':
            pass
        elif octave_type == 'single' and oct_type == 'double':
            pass
        elif octave_type == 'int64' and oct_type == 'int32':
            pass
        else:
            assert octave_type == oct_type
        if type(incoming) != in_type:
            if type(incoming) == np.int32 and in_type == np.int64:
                pass
            else:
                assert in_type(incoming) == incoming


class IncomingTest(test.TestCase):
    """Test the importing of all Octave data types, checking their type

    Uses test_datatypes.m to read in a dictionary with all Octave types
    Tests the types of all the values to make sure they were
        brought in properly.

    """
    def helper(self, base, keys, types):
        """
        Perform type checking of the values

        Parameters
        ==========
        base : dict
            Sub-dictionary we are accessing.
        keys : array-like
            List of keys to test in base.
        types : array-like
            List of expected return types for the keys.

        """
        for key, type_ in zip(keys, types):
            if not type(base[key]) == type_:
                try:
                    assert type_(base[key]) == base[key]
                except ValueError:
                    assert np.allclose(type_(base[key]), base[key])

    def test_int(self):
        """Test incoming integer types
        """
        keys = ['int8', 'int16', 'int32', 'int64',
                    'uint8', 'uint16', 'uint32', 'uint64']
        types = [np.int8, np.int16, np.int32, np.int64,
                    np.uint8, np.uint16, np.uint32, np.uint64]
        self.helper(DATA.num.int, keys, types)

    def test_floats(self):
        """Test incoming float types
        """
        keys = ['float32', 'float64', 'complex', 'complex_matrix']
        types = [np.float64, np.float64, np.complex128, np.ndarray]
        self.helper(DATA.num, keys, types)
        self.assertEqual(DATA.num.complex_matrix.dtype,
                         np.dtype('complex128'))

    def test_misc_num(self):
        """Test incoming misc numeric types
        """
        keys = ['inf', 'NaN', 'matrix', 'vector', 'column_vector', 'matrix3d',
                'matrix5d']
        types = [np.float64, np.float64, np.ndarray, np.ndarray, np.ndarray,
                 np.ndarray, np.ndarray]
        self.helper(DATA.num, keys, types)

    def test_logical(self):
        """Test incoming logical type
        """
        self.assertEqual(type(DATA.logical), np.ndarray)

    def test_string(self):
        """Test incoming string types
        """
        keys = ['basic', 'char_array', 'cell_array']
        types = [unicode, list, list]
        self.helper(DATA.string, keys, types)

    def test_struct_array(self):
        ''' Test incoming struct array types '''
        keys = ['name', 'age']
        types = [list, list]
        self.helper(DATA.struct_array, keys, types)

    def test_cell_array(self):
        ''' Test incoming cell array types '''
        keys = ['vector', 'matrix']
        types = [list, list]
        self.helper(DATA.cell, keys, types)

    def test_mixed_struct(self):
        '''Test mixed struct type
        '''
        keys = ['array', 'cell', 'scalar']
        types = [list, list, float]
        self.helper(DATA.mixed, keys, types)


class RoundtripTest(test.TestCase):
    """Test roundtrip value and type preservation between Python and Octave.

    Uses test_datatypes.m to read in a dictionary with all Octave types
    uses roundtrip.m to send each of the values out and back,
        making sure the value and the type are preserved.

    """
    def setUp(self):
        self.oc = Oct2Py()
        self.oc.addpath(os.path.dirname(__file__))

    def nested_equal(self, val1, val2):
        """Test for equality in a nested list or ndarray
        """
        if isinstance(val1, list):
            for (subval1, subval2) in zip(val1, val2):
                if isinstance(subval1, list):
                    self.nested_equal(subval1, subval2)
                elif isinstance(subval1, np.ndarray):
                    np.allclose(subval1, subval2)
                else:
                    self.assertEqual(subval1, subval2)
        elif isinstance(val1, np.ndarray):
            np.allclose(val1, np.array(val2))
        elif isinstance(val1, (str, unicode)):
            self.assertEqual(val1, val2)
        else:
            try:
                assert (np.alltrue(np.isnan(val1)) and
                        np.alltrue(np.isnan(val2)))
            except (AssertionError, NotImplementedError):
                self.assertEqual(val1, val2)

    def helper(self, outgoing, expected_type=None):
        """
        Use roundtrip.m to make sure the data goes out and back intact.

        Parameters
        ==========
        outgoing : object
            Object to send to Octave.

        """
        incoming = self.oc.roundtrip(outgoing)
        if expected_type is None:
            expected_type = type(outgoing)
        self.nested_equal(incoming, outgoing)
        try:
            self.assertEqual(type(incoming), expected_type)
        except AssertionError:
            if type(incoming) == np.float32 and expected_type == np.float64:
                pass

    def test_int(self):
        """Test roundtrip value and type preservation for integer types
        """
        for key in ['int8', 'int16', 'int32', 'int64',
                    'uint8', 'uint16', 'uint32', 'uint64']:
            self.helper(DATA.num.int[key])

    def test_float(self):
        """Test roundtrip value and type preservation for float types
        """
        for key in ['float64', 'complex', 'complex_matrix']:
            self.helper(DATA.num[key])
        self.helper(DATA.num['float32'], np.float64)

    def test_misc_num(self):
        """Test roundtrip value and type preservation for misc numeric types
        """
        for key in ['inf', 'NaN', 'matrix', 'vector', 'column_vector',
                    'matrix3d', 'matrix5d']:
            self.helper(DATA.num[key])

    def test_logical(self):
        """Test roundtrip value and type preservation for logical type
        """
        self.helper(DATA.logical)

    def test_string(self):
        """Test roundtrip value and type preservation for string types
        """
        for key in ['basic', 'cell_array']:
            self.helper(DATA.string[key])

    def test_struct_array(self):
        """Test roundtrip value and type preservation for struct array types
        """
        self.helper(DATA.struct_array['name'])
        self.helper(DATA.struct_array['age'], np.ndarray)

    def test_cell_array(self):
        """Test roundtrip value and type preservation for cell array types
        """
        for key in ['vector', 'matrix', 'array']:
            self.helper(DATA.cell[key])
        #self.helper(DATA.cell['array'], np.ndarray)

    def test_octave_origin(self):
        '''Test all of the types, originating in octave, and returning
        '''
        self.oc.run('x = test_datatypes()')
        self.oc.put('y', DATA)
        try:
            self.oc.isequaln
            func = 'isequaln'
        except Oct2PyError:
            func = 'isequalwithequalnans'
        for key in DATA.keys():
            if key != 'struct_array':
                cmd = '{0}(x.{1},y.{1})'.format(func, key)
                ret = self.oc.run(cmd)
                assert ret == 'ans =  1'


class BuiltinsTest(test.TestCase):
    """Test the exporting of standard Python data types, checking their type.

    Runs roundtrip.m and tests the types of all the values to make sure they
    were brought in properly.

    """
    def setUp(self):
        self.oc = Oct2Py()
        self.oc.addpath(os.path.dirname(__file__))

    def helper(self, outgoing, incoming=None, expected_type=None):
        """
        Uses roundtrip.m to make sure the data goes out and back intact.

        Parameters
        ==========
        outgoing : object
            Object to send to Octave
        incoming : object, optional
            Object already retreived from Octave

        """
        if incoming is None:
            incoming = self.oc.roundtrip(outgoing)
        if not expected_type:
            for out_type, _, in_type in TYPE_CONVERSIONS:
                if out_type == type(outgoing):
                    expected_type = in_type
                    break
        if not expected_type:
            expected_type = np.ndarray
        try:
            self.assertEqual(incoming, outgoing)
        except ValueError:
            assert np.allclose(np.array(incoming), np.array(outgoing))
        if type(incoming) != expected_type:
            incoming = self.oc.roundtrip(outgoing)
            assert expected_type(incoming) == incoming

    def test_dict(self):
        """Test python dictionary
        """
        test = dict(x='spam', y=[1, 2, 3])
        incoming = self.oc.roundtrip(test)
        #incoming = dict(incoming)
        for key in incoming:
            self.helper(test[key], incoming[key])

    def test_nested_dict(self):
        """Test nested python dictionary
        """
        test = dict(x=dict(y=1e3, z=[1, 2]), y='spam')
        incoming = self.oc.roundtrip(test)
        incoming = dict(incoming)
        for key in test:
            if isinstance(test[key], dict):
                for subkey in test[key]:
                    self.helper(test[key][subkey], incoming[key][subkey])
            else:
                self.helper(test[key], incoming[key])

    def test_set(self):
        """Test python set type
        """
        test = set((1, 2, 3, 3))
        incoming = self.oc.roundtrip(test)
        assert np.allclose(tuple(test), incoming)
        self.assertEqual(type(incoming), np.ndarray)

    def test_tuple(self):
        """Test python tuple type
        """
        test = tuple((1, 2, 3))
        self.helper(test, expected_type=np.ndarray)

    def test_list(self):
        """Test python list type
        """
        tests = [[1, 2], ['a', 'b']]
        self.helper(tests[0])
        self.helper(tests[1], expected_type=list)

    def test_list_of_tuples(self):
        """Test python list of tuples
        """
        test = [(1, 2), (1.5, 3.2)]
        self.helper(test)

    def test_numeric(self):
        """Test python numeric types
        """
        test = np.random.randint(1000)
        self.helper(int(test))
        self.helper(long(test))
        self.helper(float(test))
        self.helper(complex(1, 2))

    def test_string(self):
        """Test python str and unicode types
        """
        tests = ['spam', unicode('eggs')]
        for test in tests:
            self.helper(test)

    def test_nested_list(self):
        """Test python nested lists
        """
        test = [['spam', 'eggs'], ['foo ', 'bar ']]
        self.helper(test, expected_type=list)
        test = [[1, 2], [3, 4]]
        self.helper(test)
        test = [[1, 2], [3, 4, 5]]
        incoming = self.oc.roundtrip(test)
        for i in range(len(test)):
            assert np.alltrue(incoming[i] == np.array(test[i]))

    def test_bool(self):
        """Test boolean values
        """
        tests = (True, False)
        for test in tests:
            incoming = self.oc.roundtrip(test)
            self.assertEqual(incoming, test)
            self.assertEqual(incoming.dtype, np.dtype('int8'))

    def test_none(self):
        """Test sending None type
        """
        incoming = self.oc.roundtrip(None)
        assert np.isnan(incoming)


class NumpyTest(test.TestCase):
    """Check value and type preservation of Numpy arrays
    """
    codes = np.typecodes['All']
    blacklist_codes = 'V'
    blacklist_names = ['float128', 'float96', 'complex192', 'complex256']

    def setUp(self):
        self.oc = Oct2Py()
        self.oc.addpath(os.path.dirname(__file__))

    def test_scalars(self):
        """Send scalar numpy types and make sure we get the same number back.
        """
        for typecode in self.codes:
            outgoing = (np.random.randint(-255, 255) + np.random.rand(1))
            try:
                outgoing = outgoing.astype(typecode)
            except TypeError:
                continue
            if (typecode in self.blacklist_codes or
                outgoing.dtype.name in self.blacklist_names):
                self.assertRaises(Oct2PyError, self.oc.roundtrip, outgoing)
                continue
            incoming = self.oc.roundtrip(outgoing)
            if outgoing.dtype.str in ['<M8[us]', '<m8[us]']:
                outgoing = outgoing.astype(np.uint64)
            try:
                assert np.allclose(incoming, outgoing)
            except (ValueError, TypeError, NotImplementedError,
                     AssertionError):
                assert np.alltrue(np.array(incoming).astype(typecode) ==
                                   outgoing)

    def test_ndarrays(self):
        """Send ndarrays and make sure we get the same array back
        """
        for typecode in self.codes:
            for ndims in [2, 3, 4]:
                size = [np.random.randint(1, 10) for i in range(ndims)]
                outgoing = (np.random.randint(-255, 255, tuple(size)))
                try:
                    outgoing += np.random.rand(*size).astype(outgoing.dtype,
                                                             casting='unsafe')
                except TypeError:  # pragma: no cover
                    outgoing += np.random.rand(*size).astype(outgoing.dtype)
                if typecode in ['U', 'S']:
                    outgoing = [[['spam', 'eggs'], ['spam', 'eggs']],
                                [['spam', 'eggs'], ['spam', 'eggs']]]
                    outgoing = np.array(outgoing).astype(typecode)
                else:
                    try:
                        outgoing = outgoing.astype(typecode)
                    except TypeError:
                        continue
                if (typecode in self.blacklist_codes or
                     outgoing.dtype.name in self.blacklist_names):
                    self.assertRaises(Oct2PyError, self.oc.roundtrip, outgoing)
                    continue
                incoming = self.oc.roundtrip(outgoing)
                incoming = np.array(incoming)
                if outgoing.size == 1:
                    outgoing = outgoing.squeeze()
                if len(outgoing.shape) > 2 and 1 in outgoing.shape:
                    incoming = incoming.squeeze()
                    outgoing = outgoing.squeeze()
                elif incoming.size == 1:
                    incoming = incoming.squeeze()
                assert incoming.shape == outgoing.shape
                if outgoing.dtype.str in ['<M8[us]', '<m8[us]']:
                    outgoing = outgoing.astype(np.uint64)
                try:
                    assert np.allclose(incoming, outgoing)
                except (AssertionError, ValueError, TypeError,
                         NotImplementedError):
                    if 'c' in incoming.dtype.str:
                        incoming = np.abs(incoming)
                        outgoing = np.abs(outgoing)
                    assert np.alltrue(np.array(incoming).astype(typecode) ==
                                       outgoing)

    def test_sparse(self):
        '''Test roundtrip sparse matrices
        '''
        from scipy.sparse import csr_matrix, identity
        rand = np.random.rand(100, 100)
        rand = csr_matrix(rand)
        iden = identity(1000)
        for test in [rand, iden]:
            incoming, type_ = self.oc.roundtrip(test)
            assert test.shape == incoming.shape
            assert test.nnz == incoming.nnz
            assert np.allclose(test.todense(), incoming.todense())
            assert test.dtype == incoming.dtype
            assert (type_ == 'double' or type_ == 'cell')

    def test_empty(self):
        '''Test roundtrip empty matrices
        '''
        test = np.empty((100, 100))
        incoming, type_ = self.oc.roundtrip(test)
        assert test.squeeze().shape == incoming.squeeze().shape
        assert np.allclose(test[np.isfinite(test)],
                            incoming[np.isfinite(incoming)])
        assert type_ == 'double'

    def test_mat(self):
        '''Verify support for matrix type
        '''
        test = np.random.rand(1000)
        test = np.mat(test)
        incoming, type_ = self.oc.roundtrip(test)
        assert np.allclose(test, incoming)
        assert test.dtype == incoming.dtype
        assert type_ == 'double'

    def test_masked(self):
        '''Test support for masked arrays
        '''
        test = np.random.rand(100)
        test = np.ma.array(test)
        incoming, type_ = self.oc.roundtrip(test)
        assert np.allclose(test, incoming)
        assert test.dtype == incoming.dtype
        assert type_ == 'double'


class BasicUsageTest(test.TestCase):
    """Excercise the basic interface of the package
    """
    def setUp(self):
        self.oc = Oct2Py()
        self.oc.addpath(os.path.dirname(__file__))

    def test_run(self):
        """Test the run command
        """
        out = self.oc.run('y=ones(3,3)')
        desired = """y =

        1        1        1
        1        1        1
        1        1        1
"""
        self.assertEqual(out, desired)
        out = self.oc.run('x = mean([[1, 2], [3, 4]])', verbose=True)
        self.assertEqual(out, 'x =  2.5000')
        self.assertRaises(Oct2PyError, self.oc.run, '_spam')

    def test_call(self):
        """Test the call command
        """
        out = self.oc.call('ones', 1, 2)
        assert np.allclose(out, np.ones((1, 2)))
        U, S, V = self.oc.call('svd', [[1, 2], [1, 3]])
        assert np.allclose(U, ([[-0.57604844, -0.81741556],
                            [-0.81741556, 0.57604844]]))
        assert np.allclose(S,  ([[3.86432845, 0.],
                             [0., 0.25877718]]))
        assert np.allclose(V,  ([[-0.36059668, -0.93272184],
         [-0.93272184, 0.36059668]]))
        out = self.oc.call('roundtrip.m', 1)
        self.assertEqual(out, 1)
        fname = os.path.join(__file__, 'roundtrip.m')
        out = self.oc.call(fname, 1)
        self.assertEqual(out, 1)
        self.assertRaises(Oct2PyError, self.oc.call, '_spam')

    def test_put_get(self):
        """Test putting and getting values
        """
        self.oc.put('spam', [1, 2])
        out = self.oc.get('spam')
        assert np.allclose(out, np.array([1, 2]))
        self.oc.put(['spam', 'eggs'], ['foo', [1, 2, 3, 4]])
        spam, eggs = self.oc.get(['spam', 'eggs'])
        self.assertEqual(spam, 'foo')
        assert np.allclose(eggs, np.array([[1, 2, 3, 4]]))
        self.assertRaises(Oct2PyError, self.oc.put, '_spam', 1)
        self.assertRaises(Oct2PyError, self.oc.get, '_spam')

    def test_help(self):
        """Testing help command
        """
        doc = self.oc.cos.__doc__
        assert 'Compute the cosine for each element of X in radians.' in doc

    def test_dynamic(self):
        """Test the creation of a dynamic function
        """
        tests = [self.oc.zeros, self.oc.ones, self.oc.plot]
        for test in tests:
            try:
                self.assertEqual(repr(type(test)), "<type 'function'>")
            except AssertionError:
                self.assertEqual(repr(type(test)), "<class 'function'>")
        self.assertRaises(Oct2PyError, self.oc.__getattr__, 'aaldkfasd')
        self.assertRaises(Oct2PyError, self.oc.__getattr__, '_foo')
        self.assertRaises(Oct2PyError, self.oc.__getattr__, 'foo\W')

    def test_open_close(self):
        """Test opening and closing the Octave session
        """
        oct_ = Oct2Py()
        oct_.close()
        self.assertRaises(Oct2PyError, oct_.put, names=['a'],
                          var=[1.0])
        oct_.restart()
        oct_.put('a', 5)
        a = oct_.get('a')
        assert a == 5

    def test_struct(self):
        """Test Struct construct
        """
        test = Struct()
        test.spam = 'eggs'
        test.eggs.spam = 'eggs'
        self.assertEqual(test['spam'], 'eggs')
        self.assertEqual(test['eggs']['spam'], 'eggs')
        test["foo"]["bar"] = 10
        self.assertEqual(test.foo.bar, 10)
        p = pickle.dumps(test)
        test2 = pickle.loads(p)
        self.assertEqual(test2['spam'], 'eggs')
        self.assertEqual(test2['eggs']['spam'], 'eggs')
        self.assertEqual(test2.foo.bar, 10)

    def test_syntax_error(self):
        """Make sure a syntax error in Octave throws an Oct2PyError
        """
        oc = Oct2Py()
        self.assertRaises(Oct2PyError, oc._eval, "a='1")
        oc = Oct2Py()
        self.assertRaises(Oct2PyError, oc._eval, "a=1++3")

        if not oc._session.use_pexpect:
            self.assertRaises(Oct2PyError, oc._eval, "a=1")
        else:
            oc.put('a', 1)
            a = oc.get('a')
            self.assertEqual(a, 1)

    def test_octave_error(self):
        oc = Oct2Py()
        self.assertRaises(Oct2PyError, oc.run, 'a = ones2(1)')


def test_unicode_docstring():
    '''Make sure unicode docstrings in Octave functions work'''
    oc = Oct2Py()
    oc.addpath(os.path.dirname(__file__))
    help(oc.test_datatypes)


def test_context_manager():
    '''Make sure oct2py works within a context manager'''
    oc = Oct2Py()
    with oc as oc1:
        ones = oc1.ones(1)
    assert ones == np.ones(1)
    with oc as oc2:
         ones = oc2.ones(1)
    assert ones == np.ones(1)


def test_singleton_sparses():
    '''Make sure a singleton sparse matrix works'''
    import scipy.sparse
    data = scipy.sparse.csc.csc_matrix(1)
    oc = Oct2Py()
    oc.put('x', data)
    assert np.allclose(data.toarray(), oc.get('x').toarray())
    oc.put('y', [data])
    assert np.allclose(data.toarray(), oc.get('y').toarray())


def test_logging():
    # create a stringio and a handler to log to it
    def get_handler():
        sobj = StringIO()
        hdlr = logging.StreamHandler(sobj)
        hdlr.setLevel(logging.DEBUG)
        return hdlr

    oc = Oct2Py()
    hdlr = get_handler()
    oc.logger.addHandler(hdlr)

    # generate some messages (logged and not logged)
    oc.ones(1, verbose=True)

    oc.logger.setLevel(logging.DEBUG)
    oc.zeros(1)

    # check the output
    lines = hdlr.stream.getvalue().strip().split('\n')
    resp = '\n'.join(lines)
    assert 'zeros(A__)' in resp
    assert 'ans =  1' in resp
    assert lines[0].startswith('load')

    # now make an object with a desired logger
    logger = oct2py.get_log('test')
    hdlr = get_handler()
    logger.addHandler(hdlr)
    logger.setLevel(logging.INFO)
    oc2 = Oct2Py(logger=logger)

     # generate some messages (logged and not logged)
    oc2.ones(1, verbose=True)

    oc2.logger.setLevel(logging.DEBUG)
    oc2.zeros(1)

    # check the output
    lines = hdlr.stream.getvalue().strip().split('\n')
    resp = '\n'.join(lines)
    assert 'zeros(A__)' in resp
    assert 'ans =  1' in resp
    assert lines[0].startswith('load')


def test_demo():
    from oct2py import demo
    try:
        demo.demo(0.01, interactive=False)
    except AttributeError:
        demo(0.01, interactive=False)


def test_lookfor():
    assert 'cosd' in Oct2Py().lookfor('cos')


def test_remove_files():
    from oct2py.utils import _remove_temp_files
    _remove_temp_files()


def test_threads():
    from oct2py import thread_test
    thread_test()


def test_plot():
    oc = Oct2Py()
    n = oc.figure()
    oc.plot([1, 2, 3])
    oc.close_(n)


def test_narg_out():
    oc = Oct2Py()
    s = oc.svd(np.array([[1,2], [1,3]]))
    assert s.shape == (2, 1)
    U, S, V = oc.svd([[1,2], [1,3]])
    assert U.shape == S.shape == V.shape == (2, 2)


def test_help():
    help(Oct2Py())


def test_trailing_underscore():
    oc = Oct2Py()
    x = oc.ones_()
    assert np.allclose(x, np.ones(1))


def test_using_closed_session():
    oc = Oct2Py()
    oc.close()
    test.assert_raises(Oct2PyError, oc.call, 'ones')


def test_keyboard():

    oc = Oct2Py()

    oc._eval('a=1')

    stdin = sys.stdin
    stdout = sys.stdout
    output = StringIO()
    sys.stdin = StringIO('a\nreturn')
    oc._session.stdout = output
    try:
        oc.keyboard(timeout=3)
    except Oct2PyError as e:
        if 'session timed out' in str(e).lower():
            return
    sys.stdin.flush()
    sys.stdin = stdin
    oc._session.stdout = stdout

    out = output.getvalue()
    assert 'Entering Octave Debug Prompt...' in out
    assert 'a =  1' in out


def test_func_without_docstring():
    oc = Oct2Py()
    oc.addpath(os.path.dirname(__file__))
    out = oc.test_nodocstring(5)
    assert out == 5
    assert 'user-defined function' in oc.test_nodocstring.__doc__
    assert os.path.dirname(__file__) in oc.test_nodocstring.__doc__


def test_func_noexist():
    oc = Oct2Py()
    test.assert_raises(Oct2PyError, oc.call, 'oct2py_dummy')


def test_timeout():
    oc = Oct2Py(timeout=2)
    oc.sleep(2.1, timeout=5)
    test.assert_raises(Oct2PyError, oc.sleep, 3)


def test_call_path():
    oc = Oct2Py()
    oc.addpath(os.path.dirname(__file__))
    DATA = oc.call('test_datatypes.m')
    assert DATA.string.basic == 'spam'

    oc = Oct2Py()
    path = os.path.join(os.path.dirname(__file__), 'test_datatypes.m')
    DATA = oc.call(path)
    assert DATA.string.basic == 'spam'
    

def test_long_variable_name():
    oc = Oct2Py()
    name = 'this_variable_name_is_over_32_char'
    oc.put(name, 1)
    x = oc.get(name)
    assert x == 1


if __name__ == '__main__':  # pragma: no cover
    print('oct2py test')
    print('*' * 20)
    test.run_module_suite()



# -*- coding: utf-8 -*-
"""
Oct2Py is a means to seamlessly call M-files and GNU Octave functions from Python.
It manages the Octave session for you, sharing data behind the scenes using
MAT files.  Usage is as simple as:

.. code-block:: python

    >>> import oct2py
    >>> oc = oct2py.Oct2Py() 
    >>> x = oc.zeros(3,3)
    >>> print x, x.dtype
    [[ 0.  0.  0.]
     [ 0.  0.  0.]
     [ 0.  0.  0.]] float64

If you want to run legacy m-files, do not have MATLAB(TM), and do not fully
trust a code translator, this is your library.  
"""


__title__ = 'oct2py'
__version__ = '1.4.0'
__author__ = 'Steven Silvester'
__license__ = 'MIT'
__copyright__ = 'Copyright 2014 Steven Silvester'
__all__ = ['Oct2Py', 'Oct2PyError', 'octave', 'Struct', 'demo', 'speed_test',
           'thread_test', '__version__', 'get_log']


import imp
import functools
import os

from .session import Oct2Py, Oct2PyError
from .utils import Struct, get_log
from .demo import demo
from .speed_check import speed_test
from .thread_check import thread_test


try:
    octave = Oct2Py()
except Oct2PyError as e:
    print(e)

# clean up namespace
del functools, imp, os
try:
    del session, utils, speed_check, thread_check
except NameError:  # pragma: no cover
    pass

